/*
  led-ulog
  Matlib  2010-02-07
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/netfilter_ipv4/ipt_ULOG.h> 


#include <arguments.h>
#include <proto/led.c>


#define CHECK_ERROR(__COND__,__FUN__) \
  if (__COND__) \
  { \
    perror (__FUN__); \
    exit (EXIT_FAILURE); \
  }
#define CHECK_ERROR_INT(__COND__,...) \
  if (__COND__) \
  { \
    fprintf (stderr, __VA_ARGS__); \
    exit (EXIT_FAILURE); \
  }
#define LED_ERROR_CHK(__COND__) \
  CHECK_ERROR_INT (__COND__, "%s,%s,%ld: %s\n", \
     ARG_HOST, ARG_PORT, ARG_LED, led_strerror (e, buffer, ARG_BSIZ))


/* create_nl_socket (<group mask>)
   Creates new netlink socket and binds to groups <group mask>. */
  
#define HDRLEN (sizeof (struct nlmsghdr) + sizeof (struct ulog_packet_msg) - sizeof (void *))

int create_nl_socket (int group_mask)
{
  struct sockaddr_nl sa;
  int fd;
  
  fd = socket (PF_NETLINK, SOCK_RAW, NETLINK_NFLOG);
  CHECK_ERROR (fd == -1, "socket");
  bzero (&sa, sizeof (sa));
  sa.nl_family = AF_NETLINK;
  sa.nl_groups = group_mask;
  CHECK_ERROR (bind (fd, (struct sockaddr *) &sa, sizeof (sa)) != 0, "bind");
  return fd;
}
  
void set_io_modes (int fd)
{
  int flags;  
  flags = fcntl (fd, F_GETFL, 0);
  CHECK_ERROR (flags == -1, "fcntl (F_GETFL)");
  CHECK_ERROR (fcntl (fd, F_SETFL, flags | O_ASYNC | O_NONBLOCK) == -1, 
     "fcntl (F_SETFL)");
  CHECK_ERROR (fcntl (fd, F_SETOWN, getpid ()), "fcntl (F_SETOWN)");
}


/* connect_to_led_server (<host>, <port>)
   Universal IPv6/IPv4 connection routine */

int connect_to_led_server (char *host, char *port)
{
  struct addrinfo hints;
  struct addrinfo *ai, *i;
  int e, en, s;
  char hbuffer [1024], pbuffer [128];
  
  bzero (&hints, sizeof (struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = 0;
  hints.ai_protocol = 0;
  e = getaddrinfo (host, port, &hints, &ai);
  CHECK_ERROR_INT (e != 0, "%s:%s: getaddrinfo: %s\n", host, port, 
     gai_strerror (e));

  for (i = ai; i != NULL; i = i -> ai_next)
  {
    s = socket (i -> ai_family, i -> ai_socktype, i -> ai_protocol);
    if (s == -1)
    {
      fprintf (stderr, "socket (family 0x%X, type %d, proto %d): %s", 
         i -> ai_family, i -> ai_socktype, i -> ai_protocol, strerror (errno));
      continue;
    }
    if (connect (s, i -> ai_addr, i -> ai_addrlen) != -1)
    {
      freeaddrinfo (ai);
      return s;
    }
    en = errno;
    e = getnameinfo ((struct sockaddr *) i -> ai_addr, i -> ai_addrlen, 
       hbuffer, sizeof (hbuffer), pbuffer, sizeof (pbuffer), NI_NUMERICHOST);
    if (e != 0)
      fprintf (stderr, "%s:%s: connect: %s, additionally\ngetnameinfo: %s\n", 
         host, port, strerror (en), gai_strerror (e));
    else
      fprintf (stderr, "%s:%s: connect: %s\n", hbuffer, pbuffer, strerror (en));
    close (s);
  }
  exit (EXIT_FAILURE);
}


/* signal_ignore
   Ignore signal handler: linux SIG_IGN seems not to work as expected
   (behaves the same as blocked signal). We need to define our own function
   then. */
void signal_ignore (int signum)
{
  signal (signum, signal_ignore);
}

/* signal_pipe
   Prints a message after receiving SIGPIPE to save you from additional
   hair loss. */
void signal_pipe (int signum)
{
  fprintf (stderr, "SIGPIPE received, possibly due to LED server "
     "closing connection\n");
  fflush (stderr);
  signal (signum, SIG_DFL);
  raise (signum);
}

/* unblock_signals */
void unblock_signals ()
{
  sigset_t ss;
  sigemptyset (&ss);
  sigaddset (&ss, SIGIO);
  sigaddset (&ss, SIGPIPE);
  CHECK_ERROR (sigprocmask (SIG_UNBLOCK, &ss, NULL), "sigprocmask (SIG_UNBLOCK)");
  sigemptyset (&ss);
  sigaddset (&ss, SIGHUP);
  CHECK_ERROR (sigprocmask (SIG_BLOCK, &ss, NULL), "sigprocmask (SIG_BLOCK)");
}

/* daemon_mode
   Switch to background operation */
void daemon_mode ()
{
  pid_t pid;

  pid = fork ();
  CHECK_ERROR (pid == -1, "fork");
  if (pid != 0)
    exit (EXIT_SUCCESS);
  fclose (stdin);
  fclose (stdout);
}




struct Argument arguments [] =
{
  {'B', "buffer-size", argument_int,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) 2048},
  {'f', "foreground",  argument_bool,   ARGUMENT_F_ONCE,                  NULL},
  {'b', "link-speed",  argument_int,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) 1024},
  {'g', "group-mask",  argument_int,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) -1},
  {'h', "host",        argument_string, ARGUMENT_F_ONCE | ARGUMENT_F_PAR, "::1"},
  {'p', "port",        argument_string, ARGUMENT_F_ONCE | ARGUMENT_F_PAR, "335"},
  {'l', "led",         argument_int,    ARGUMENT_F_ONCE | ARGUMENT_F_PAR, (void *) 0},
  {}
};
#define ARG_BSIZ ((long int) arguments [0].value)
#define ARG_FGND (NULL !=    arguments [1].value)
#define ARG_KBPS ((long int) arguments [2].value)
#define ARG_GMSK ((long int) arguments [3].value)
#define ARG_HOST ((char *)   arguments [4].value)
#define ARG_PORT ((char *)   arguments [5].value)
#define ARG_LED  ((long int) arguments [6].value)





int main (int argc, char **argv)
{
  int e, i, nl, ls;
  ssize_t s, rc;
  char *buffer;
  struct timespec ts;

  buffer = malloc (ARG_BSIZ);
  CHECK_ERROR (buffer == NULL, "malloc");
  unblock_signals ();
  signal (SIGIO, signal_ignore);
  signal (SIGPIPE, signal_pipe);
  
  i = arguments_parse ("led-packet", "1 (2010-02-09)", "", 
     "Blink connected LED on packets captured with ULOG", argc, argv, 
     arguments);
  nl = create_nl_socket (ARG_GMSK);
  ls = connect_to_led_server (ARG_HOST, ARG_PORT);
  LED_ERROR_CHK ((e = led_add (ls, ARG_LED)) != 0)
  LED_ERROR_CHK ((e = led_set (ls, 0)) != 0);

  ts.tv_sec = 0;
  i = 8e6 / ARG_KBPS;  // nanosleep = (8 bits / b kbits/s) * 10e6  [ns]
  if (!ARG_FGND)
    daemon_mode ();
  set_io_modes (nl);
  
  s = 0;
loop:
  rc = recv (nl, buffer, ARG_BSIZ, MSG_WAITALL);
  if (rc == 0)
  {
    fprintf (stderr, "Netlink connection closed");
    exit (EXIT_FAILURE);
  }
  if (rc == -1)
  {
    if (errno != EWOULDBLOCK)
      { CHECK_ERROR (errno != EINTR,  "recv") }
    else
    {
      if (s < HDRLEN)
        sleep (1);
      else
      {
        LED_ERROR_CHK ((e = led_set (ls, 100)) != 0);
        ts.tv_nsec = (s - HDRLEN) * i;
        nanosleep (&ts, NULL);
        LED_ERROR_CHK ((e = led_set (ls, 0)) != 0);
        s = 0;
      }
    }
  }
  else
    s = rc;
  goto loop;
}
